# FixDCOM.ps1 - 2017-08-04 - Ingo Böttcher
# This script is changing access rights to registry keys and DCOM objects to solve DCOM event 10016
# It is largely based on existing scripts from the net.
# 
# Sources:
# http://www.leeholmes.com/blog/2010/09/24/adjusting-token-privileges-in-powershell/
# https://gallery.technet.microsoft.com/Set-DCOM-ACL-with-650fa48d
# https://powertoe.wordpress.com/2010/08/28/controlling-registry-acl-permissions-with-powershell/
#
# Usage:
#
# From eventviewer DCOM event 10016 please note affected CLSID and AppID and run:
#
# .\FixDCOM.ps1 "<CLSID>" "<AppID>"
#
# For example:
#
# .\FixDCOM.ps1 "{6B3B8D23-FA8D-40B9-8DBD-B950333E2C52}" "{4839DDB7-58C2-48F5-8283-E1D1807D0D7D}"
#

#Requires -RunAsAdministrator
param (
  [Parameter(Mandatory=$True)][string]$ProblemCLSID, 
  [Parameter(Mandatory=$True)][string]$ProblemAppID
)


$LocalAdminSID = "S-1-5-32-544"  # local Administrators
$LocalServiceSID = "S-1-5-19" # Local Service

function enable-privilege {
 param(
 ## The privilege to adjust. This set is taken from
 ## http://msdn.microsoft.com/en-us/library/bb530716(VS.85).aspx
 [ValidateSet(
 "SeAssignPrimaryTokenPrivilege", "SeAuditPrivilege", "SeBackupPrivilege",
 "SeChangeNotifyPrivilege", "SeCreateGlobalPrivilege", "SeCreatePagefilePrivilege",
 "SeCreatePermanentPrivilege", "SeCreateSymbolicLinkPrivilege", "SeCreateTokenPrivilege",
 "SeDebugPrivilege", "SeEnableDelegationPrivilege", "SeImpersonatePrivilege", "SeIncreaseBasePriorityPrivilege",
 "SeIncreaseQuotaPrivilege", "SeIncreaseWorkingSetPrivilege", "SeLoadDriverPrivilege",
 "SeLockMemoryPrivilege", "SeMachineAccountPrivilege", "SeManageVolumePrivilege",
 "SeProfileSingleProcessPrivilege", "SeRelabelPrivilege", "SeRemoteShutdownPrivilege",
 "SeRestorePrivilege", "SeSecurityPrivilege", "SeShutdownPrivilege", "SeSyncAgentPrivilege",
 "SeSystemEnvironmentPrivilege", "SeSystemProfilePrivilege", "SeSystemtimePrivilege",
 "SeTakeOwnershipPrivilege", "SeTcbPrivilege", "SeTimeZonePrivilege", "SeTrustedCredManAccessPrivilege",
 "SeUndockPrivilege", "SeUnsolicitedInputPrivilege")]
 $Privilege,
 ## The process on which to adjust the privilege. Defaults to the current process.
 $ProcessId = $pid,
 ## Switch to disable the privilege, rather than enable it.
 [Switch] $Disable
 )
 
## Taken from P/Invoke.NET with minor adjustments.
 $definition = @'
 using System;
 using System.Runtime.InteropServices;
 
public class AdjPriv
{
[DllImport("advapi32.dll", ExactSpelling = true, SetLastError = true)]
internal static extern bool AdjustTokenPrivileges(IntPtr htok, bool disall,
ref TokPriv1Luid newst, int len, IntPtr prev, IntPtr relen);
 
[DllImport("advapi32.dll", ExactSpelling = true, SetLastError = true)]
internal static extern bool OpenProcessToken(IntPtr h, int acc, ref IntPtr phtok);
[DllImport("advapi32.dll", SetLastError = true)]
internal static extern bool LookupPrivilegeValue(string host, string name, ref long pluid);
[StructLayout(LayoutKind.Sequential, Pack = 1)]
internal struct TokPriv1Luid
{
public int Count;
public long Luid;
public int Attr;
}
 
internal const int SE_PRIVILEGE_ENABLED = 0x00000002;
internal const int SE_PRIVILEGE_DISABLED = 0x00000000;
internal const int TOKEN_QUERY = 0x00000008;
internal const int TOKEN_ADJUST_PRIVILEGES = 0x00000020;
public static bool EnablePrivilege(long processHandle, string privilege, bool disable)
{
bool retVal;
TokPriv1Luid tp;
IntPtr hproc = new IntPtr(processHandle);
IntPtr htok = IntPtr.Zero;
retVal = OpenProcessToken(hproc, TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, ref htok);
tp.Count = 1;
tp.Luid = 0;
if(disable)
{
tp.Attr = SE_PRIVILEGE_DISABLED;
}
else
{
tp.Attr = SE_PRIVILEGE_ENABLED;
}
retVal = LookupPrivilegeValue(null, privilege, ref tp.Luid);
retVal = AdjustTokenPrivileges(htok, false, ref tp, 0, IntPtr.Zero, IntPtr.Zero);
return retVal;
}
}
'@
 
$processHandle = (Get-Process -id $ProcessId).Handle
$type = Add-Type $definition -PassThru
$type[0]::EnablePrivilege($processHandle, $Privilege, $Disable)
}
 
Function ResetAccess($CLSID, $AppID)
{
  $CLSIDKey = "CLSID\"+$CLSID
  $AppIDKey = "Software\Classes\AppID\"+$AppID

  # Get local admins name, since RegistryAccessRule does not work with SID
  $AdmSID = New-Object System.Security.Principal.SecurityIdentifier ($LocalAdminSID)
  $AdmName = $AdmSID.Translate([System.Security.Principal.NTAccount])
   
  # Take Ownership of CLSID Key
  $key = [Microsoft.Win32.Registry]::ClassesRoot.OpenSubKey($CLSIDKey,[Microsoft.Win32.RegistryKeyPermissionCheck]::ReadWriteSubTree,[System.Security.AccessControl.RegistryRights]::TakeOwnership)
  $acl = $key.GetAccessControl([System.Security.AccessControl.AccessControlSections]::None)
  $acl.SetOwner((New-Object System.Security.Principal.SecurityIdentifier($LocalAdminSID))) 
  $key.SetAccessControl($acl)

  # Set Access Rights
  $acl = $key.GetAccessControl()
  $rule = New-Object System.Security.AccessControl.RegistryAccessRule ($AdmName,"FullControl","ObjectInherit,ContainerInherit","None","Allow")
  $acl.SetAccessRule($rule)
  $key.SetAccessControl($acl)

  $key.Close()
  
  # Take Ownership of AppID Key
  $key = [Microsoft.Win32.Registry]::LocalMachine.OpenSubKey($AppIDKey,[Microsoft.Win32.RegistryKeyPermissionCheck]::ReadWriteSubTree,[System.Security.AccessControl.RegistryRights]::TakeOwnership)
  $acl = $key.GetAccessControl([System.Security.AccessControl.AccessControlSections]::None)
  $acl.SetOwner((New-Object System.Security.Principal.SecurityIdentifier($LocalAdminSID))) 
  $key.SetAccessControl($acl)

  # Set Access Rights
  $acl = $key.GetAccessControl()
  $rule = New-Object System.Security.AccessControl.RegistryAccessRule ($AdmName,"FullControl","ObjectInherit,ContainerInherit","None","Allow")
  $acl.SetAccessRule($rule)
  $key.SetAccessControl($acl)

  $key.Close()
  
}

Function SetDACL($Caption)
{  
  Write-Host "Setting Local Service rights for $Caption"
  $filter = 'caption="'+$Caption+'"'
  $wmi = Get-WmiObject -Class Win32_DCOMApplicationSetting -Filter $filter -EnableAllPrivileges

  $descL  = $wmi.GetLaunchSecurityDescriptor().descriptor
  $descA  = $wmi.GetAccessSecurityDescriptor().descriptor
  $descC  = $wmi.GetConfigurationSecurityDescriptor().descriptor

  $trusteeObj = ([wmiclass]'Win32_Trustee').psbase.CreateInstance()
  $trusteeObj.SIDString = $LocalServiceSID

  $ace = ([wmiclass]'Win32_ACE').psbase.CreateInstance()
  $ace.AccessMask = 31
  $ace.trustee = $trusteeObj

  $descL.DACL += [System.Management.ManagementBaseObject]$ace
  $descA.DACL += [System.Management.ManagementBaseObject]$ace
  $descC.DACL += [System.Management.ManagementBaseObject]$ace

  $err = $wmi.SetLaunchSecurityDescriptor($descL)
  $err = $wmi.SetAccessSecurityDescriptor($descA)
  $err = $wmi.SetAccessSecurityDescriptor($descC)
}

Function GetDCOMName($GUID)
{
  $key = 'HKLM:\SOFTWARE\Classes\AppID\'+$GUID
  return (Get-ItemProperty -LiteralPath $key).'(default)'
}

Write-Host "This script will now try to fix the access rights in registry and DCOM."

$err = Enable-Privilege SeTakeOwnershipPrivilege
ResetAccess $ProblemCLSID $ProblemAppID
$DCOMName = GetDCOMName $ProblemAppID
SetDACL $DCOMName

Write-Host "Done."